unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, TeeGDIPlus, TeEngine, ExtCtrls, TeeProcs, Chart, Series, TeeTools,
  StdCtrls, Math, EditChar;

type
  TIntegerArray = array of Integer;
  TDoubleArray = array of Double;

  TForm1 = class(TForm)
    Chart1: TChart;
    Panel1: TPanel;
    GBOptions: TGroupBox;
    CBPointersHoriz: TCheckBox;
    CBSegmentsHoriz: TCheckBox;
    CBInterpolateHoriz: TCheckBox;
    Button1: TButton;
    GroupBox1: TGroupBox;
    CBPointersVert: TCheckBox;
    CBSegmentsVert: TCheckBox;
    CBInterpolateVert: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure CBPointersHorizClick(Sender: TObject);
    procedure CBSegmentsHorizClick(Sender: TObject);
    procedure CBInterpolateHorizClick(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure CBPointersVertClick(Sender: TObject);
    procedure CBSegmentsVertClick(Sender: TObject);
    procedure CBInterpolateVertClick(Sender: TObject);
  private
    { Private declarations }
    function InterpGetPointerStyle(Sender:TChartSeries; ValueIndex:Integer): TSeriesPointerStyle;
    procedure checkColors;
    function appendDouble(value: Double; toArray: TDoubleArray): TDoubleArray;
    procedure recalcInterpolations;
    procedure ColorLineDrag(Sender:TColorLineTool);
    function InterpolateSeriesHoriz(Series: TChartSeries; YValue: Double): TDoubleArray;
    function SearchCrossingPoints(Values: TChartValues; ACount: Integer; YValue: double): TIntegerArray;
    function InterpolateSeriesVert(Series: TChartSeries; XValue: Double): Double;
    function BinarySearch(Values: TChartValues; ACount: Integer; XValue: double): Integer;
    procedure BeforeDrawSeries(Sender: TObject);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

var sourceLine, interpLine: TLineSeries;
    colorLineHoriz, colorLineVert: TColorLineTool;
    interpPointsHoriz: TDoubleArray;
    interpPointVert: Double;

procedure TForm1.FormCreate(Sender: TObject);
begin
  Chart1.View3D:=false;
  Chart1.Legend.Visible:=false;

  sourceLine:=Chart1.AddSeries(TLineSeries) as TLineSeries;
  with sourceLine do
  begin
    Pointer.Visible:=true;
    ColorEachLine:=true;

    FillSampleValues;
  end;

  interpLine:=CloneChartSeries(sourceLine) as TLineSeries;
  interpLine.OnGetPointerStyle:=InterpGetPointerStyle;
  Chart1.OnBeforeDrawSeries:=BeforeDrawSeries;

  sourceLine.Visible:=false;

  colorLineHoriz:=Chart1.Tools.Add(TColorLineTool) as TColorLineTool;
  with colorLineHoriz do
  begin
    Axis:=Chart1.Axes.Left;
    Value:=sourceLine.YValues.MinValue + (sourceLine.YValues.MaxValue - sourceLine.YValues.MinValue) / 2;
    OnDragLine:=ColorLineDrag;
    Pen.Width:=2;
    Pen.Color:=clRed;
  end;

  colorLineVert:=Chart1.Tools.Add(TColorLineTool) as TColorLineTool;
  with colorLineVert do
  begin
    Axis:=Chart1.Axes.Bottom;
    Value:=sourceLine.XValues.MinValue + (sourceLine.XValues.MaxValue - sourceLine.XValues.MinValue) / 2;
    OnDragLine:=ColorLineDrag;
    Pen.Width:=2;
    Pen.Style:=psDot;
  end;

  CBInterpolateHoriz.Enabled:=CBSegmentsHoriz.Checked;
  CBInterpolateVert.Enabled:=CBSegmentsVert.Checked;
  checkColors;
end;

procedure TForm1.checkColors;
var pointerColor, segmentColor: TColor;
    i: Integer;
begin
  Chart1.AutoRepaint:=false;

  if CBPointersHoriz.Checked then
    pointerColor:=clRed
  else
    pointerColor:=sourceLine.Color;

  if CBSegmentsHoriz.Checked then
    segmentColor:=clRed
  else
    segmentColor:=sourceLine.Color;

  for i:=0 to interpLine.Count-1 do
    if (colorLineHoriz<>nil) then
    begin
      interpLine.Pointer.Items[i].Color:=sourceLine.Color;
      interpLine.ValueColor[i]:=sourceLine.Color;

      if interpLine.YValue[i]<colorLineHoriz.Value then
      begin
        interpLine.Pointer.Items[i].Color:=pointerColor;
        interpLine.ValueColor[i]:=segmentColor;
      end
      else
      if interpLine.YValue[i]=colorLineHoriz.Value then
      begin
        if (i-1>-1) and (interpLine.YValue[i-1]<=colorLineHoriz.Value) then
        begin
          interpLine.Pointer.Items[i].Color:=pointerColor;
          interpLine.ValueColor[i]:=segmentColor;
        end;
      end;
    end;

  Chart1.AutoRepaint:=true;
  Chart1.Draw;
end;

function TForm1.InterpGetPointerStyle(Sender:TChartSeries; ValueIndex:Integer): TSeriesPointerStyle;
begin
  result:=psRectangle;

  if (CBPointersVert.Checked) and
     (ValueIndex>-1) and (ValueIndex<Sender.Count) then
  begin
    if (Sender.XValue[ValueIndex] < colorLineVert.Value) then
      result:=psCircle;

    if (Sender.YValue[ValueIndex] = colorLineHoriz.Value) then
      result:=psNothing;
  end;

  if CBInterpolateVert.Enabled and CBInterpolateVert.Checked and
     (Sender.XValue[ValueIndex] = colorLineVert.Value) and
     (sourceLine.XValues.Locate(colorLineVert.Value)=-1) then
      result:=psNothing;

  interpLine.LinePen.Style:=psSolid;

  if (CBSegmentsVert.Checked) and
     (ValueIndex+1>-1) and (ValueIndex+1<Sender.Count) then
  begin
    if Sender.XValue[ValueIndex+1] < colorLineVert.Value then
      interpLine.LinePen.Style:=psDot;
  end;
end;

procedure TForm1.BeforeDrawSeries(Sender: TObject);
begin
  if (CBSegmentsVert.Checked) and (interpLine<>nil) and
     (interpLine.Count>=1) and (interpLine.XValue[0] < colorLineVert.Value) then
    interpLine.LinePen.Style:=psDot
  else
    interpLine.LinePen.Style:=psSolid;
end;

procedure TForm1.recalcInterpolations;
var i: Integer;
begin
  interpLine.Clear;
  interpLine.DataSource:=sourceLine;
  interpLine.CheckDataSource;

  if CBSegmentsHoriz.Checked and CBInterpolateHoriz.Checked then
  begin
    interpPointsHoriz:=InterpolateSeriesHoriz(sourceLine, colorLineHoriz.Value);

    for i:=0 to High(interpPointsHoriz) do
      interpLine.AddXY(interpPointsHoriz[i], colorLineHoriz.Value);
  end;

  if CBSegmentsVert.Checked and CBInterpolateVert.Checked then
  begin
    interpPointVert:=InterpolateSeriesVert(sourceLine, colorLineVert.Value);
    if interpPointVert<>MaxInt then
      interpLine.AddXY(colorLineVert.Value, interpPointVert);
  end;
end;

function TForm1.appendDouble(value: Double; toArray: TDoubleArray): TDoubleArray;
var i: Integer;
begin
  SetLength(result, length(toArray)+1);
  for i:=0 to High(toArray) do
    result[i]:=toArray[i];

  result[length(toArray)]:=value;
end;

function TForm1.InterpolateSeriesHoriz(Series: TChartSeries; YValue: Double): TDoubleArray;
var
  crossIndex: TIntegerArray;
  i: Integer;
  dx, dy: Double;
begin
  SetLength(result, 0);

  crossIndex:=SearchCrossingPoints(Series.YValues.Value, Series.Count, YValue);

  for i:=0 to High(crossIndex) do
  begin
    // safeguard
    if (crossIndex[i] < 1) then crossIndex[i] := 1
    else if (crossIndex[i] >= Series.Count) then crossIndex[i]:=Series.Count-1;

    // x = (y-y1)*(x2-x1)/(y2-y1)+x1
    dx := Series.XValues.Value[crossIndex[i]] - Series.XValues.Value[crossIndex[i] - 1];
    dy := Series.YValues.Value[crossIndex[i]] - Series.YValues.Value[crossIndex[i] - 1];

    if (dx <> 0) then
      Result:=appendDouble(((YValue - Series.YValues.Value[crossIndex[i] - 1]) * dx / dy) + Series.XValues.Value[crossIndex[i] - 1], result);
  end;
end;

function TForm1.SearchCrossingPoints(Values: TChartValues; ACount: Integer; YValue: double): TIntegerArray;

  function appendInt(value: Integer; toArray: TIntegerArray): TIntegerArray;
  var i: Integer;
  begin
    SetLength(result, length(toArray)+1);
    for i:=0 to High(toArray) do
      result[i]:=toArray[i];

    result[length(toArray)]:=value;
  end;

var i: Integer;
begin
  SetLength(result, 0);

  for i:=0 to ACount-2 do
  begin
    if ((Values[i]>YValue) and (Values[i+1]<YValue)) or
       ((Values[i]<YValue) and (Values[i+1]>YValue)) then

      result:=appendInt(i+1, result);
  end;
end;

function TForm1.InterpolateSeriesVert(Series: TChartSeries; XValue: Double): Double;
var
  Index: Integer;
  dx, dy: Double;
  FirstIdx, LastIdx: Integer;

begin
  Index:=BinarySearch(Series.XValues.Value, Series.Count, XValue);

  // safeguard
  if (Index < 1) or (Index >= Series.Count) or (Series.XValue[Index] = XValue) then
  begin
    Result:=MaxInt;
    exit;
  end;

  // y=(y2-y1)/(x2-x1)*(x-x1)+y1
  dx := Series.XValues.Value[Index] - Series.XValues.Value[Index - 1];
  dy := Series.YValues.Value[Index] - Series.YValues.Value[Index - 1];

  if (dx <> 0) then
    Result := (dy / dx) * (XValue - Series.XValues.Value[Index - 1]) + Series.YValues.Value[Index - 1];
end;

function TForm1.BinarySearch(Values: TChartValues; ACount: Integer; XValue: double): Integer;
var iMin, iMax, iMid: Integer;
begin
  iMin:=0;
  iMax:=ACount-1;

  while iMax >= iMin do
  begin
    iMid:=(iMax + iMin) shr 1;

    if SameValue(Values[iMid], XValue) then
    begin
      Result:=iMid;
      exit;
    end else
    if Values[iMid] < XValue then
      iMin:=iMid + 1
    else
      iMax:=iMid - 1;
  end;

  Result:=iMin;
end;

procedure TForm1.CBPointersHorizClick(Sender: TObject);
begin
  checkColors;
end;

procedure TForm1.CBPointersVertClick(Sender: TObject);
begin
  checkColors;
end;

procedure TForm1.CBSegmentsHorizClick(Sender: TObject);
begin
  CBInterpolateHoriz.Enabled:=CBSegmentsHoriz.Checked;

  recalcInterpolations;
  checkColors;
end;

procedure TForm1.CBSegmentsVertClick(Sender: TObject);
begin
  CBInterpolateVert.Enabled:=CBSegmentsVert.Checked;

  recalcInterpolations;
  checkColors;
end;

procedure TForm1.CBInterpolateHorizClick(Sender: TObject);
begin
  recalcInterpolations;
  checkColors;
end;

procedure TForm1.CBInterpolateVertClick(Sender: TObject);
begin
  recalcInterpolations;
  checkColors;
end;

procedure TForm1.ColorLineDrag(Sender:TColorLineTool);
begin
  if (CBInterpolateHoriz.Enabled and CBInterpolateHoriz.Checked) or
     (CBInterpolateVert.Enabled and CBInterpolateVert.Checked) then
    recalcInterpolations;

  checkColors;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  EditChart(Self, Chart1);
end;

end.
